package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.EncryptedSubmissionPayload;
import dev.fitko.fitconnect.client.sender.model.SubmissionPayload;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.UUID;
import java.util.regex.Pattern;

public class ValidDataGuard {

    private static final Logger LOGGER = LoggerFactory.getLogger(ValidDataGuard.class);
    private static final Pattern LEIKA_KEY_PATTERN = Pattern.compile("^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$");
    private final Sender sender;

    public ValidDataGuard(final Sender sender) {
        this.sender = sender;
    }

    public void ensureValidDataPayload(final EncryptedSubmissionPayload encryptedSubmissionPayload) {
        if (encryptedSubmissionPayload == null) {
            throw new IllegalArgumentException("Encrypted payload must not be null.");
        }
        if (encryptedSubmissionPayload.getEncryptedMetadata() == null) {
            throw new IllegalArgumentException("Encrypted metadata must not be null.");
        }
        testDefaults(encryptedSubmissionPayload.getDestinationId(), encryptedSubmissionPayload.getEncryptedData(), encryptedSubmissionPayload.getServiceType());
    }

    public void ensureValidDataPayload(final SubmissionPayload submissionPayload) {
        if (submissionPayload == null) {
            throw new IllegalArgumentException("Payload must not be null.");
        }
        if (isDataFormatInvalid(submissionPayload)) {
            throw new IllegalArgumentException("Data format is invalid, please provide well-formed data.");
        }
        testDefaults(submissionPayload.getDestinationId(), submissionPayload.getData(), submissionPayload.getServiceType());
    }

    public void testDefaults(final UUID destinationId, final String data, final ServiceType serviceType) {
        if (destinationId == null) {
            throw new IllegalArgumentException("DestinationId is mandatory, but was null.");
        } else if (data == null) {
            throw new IllegalArgumentException("Data is mandatory, but was null.");
        } else if (serviceType == null) {
            throw new IllegalArgumentException("ServiceType is mandatory, but was null.");
        } else if (serviceType.getIdentifier() == null) {
            throw new IllegalArgumentException("Leika key is mandatory, but was null.");
        } else if (noValidLeikaKeyPattern(serviceType.getIdentifier())) {
            throw new IllegalArgumentException("LeikaKey has invalid format, please follow: ^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$.");
        }
    }

    private boolean isDataFormatInvalid(final SubmissionPayload submissionPayload) {

        final MimeType dataMimeType = submissionPayload.getDataMimeType();
        if (dataMimeType.equals(MimeType.APPLICATION_JSON)) {
            final ValidationResult validationResult = sender.validateJsonFormat(submissionPayload.getData());
            if (validationResult.hasError()) {
                LOGGER.error("Data is not in expected json format, please provide valid json {}", validationResult.getError().getMessage());
                return true;
            }
        } else if (dataMimeType.equals(MimeType.APPLICATION_XML)) {
            final ValidationResult validationResult = sender.validateXmlFormat(submissionPayload.getData());
            if (validationResult.hasError()) {
                LOGGER.error("Data is not in expected xml format, please provide valid xml {}", validationResult.getError().getMessage());
                return true;
            }
        }
        return false;
    }

    private boolean noValidLeikaKeyPattern(final String leikaKey) {
        if (leikaKey.length() < 7 || leikaKey.length() > 255) {
            return true;
        }
        return !LEIKA_KEY_PATTERN.matcher(leikaKey).matches();
    }
}
